<?php

    /** @noinspection PhpMissingFieldTypeInspection */

    namespace TamerLib\Classes;

    use Exception;
    use LogLib\Log;
    use Redis;
    use RedisException;
    use TamerLib\Enums\JobStatus;
    use TamerLib\Exceptions\JobManagerException;
    use TamerLib\Exceptions\JobNotFoundException;
    use TamerLib\Exceptions\ConnectionException;
    use TamerLib\Exceptions\TimeoutException;
    use TamerLib\Objects\JobPacket;
    use TamerLib\Objects\ServerConfiguration;
    use TamerLib\Objects\WorkerConfiguration;
    use TamerLib\tm;
    use Throwable;

    class JobManager
    {
        /**
         * @var ServerConfiguration
         */
        private $server_configuration;

        /**
         * @var Redis
         */
        private $redis_client;

        /**
         * @var int|null
         */
        private $last_connect;

        /**
         * @var int
         */
        private $ping_count;

        /**
         * JobManager constructor.
         *
         * @param ServerConfiguration $serverConfiguration
         */
        public function __construct(ServerConfiguration $serverConfiguration)
        {
            $this->server_configuration = $serverConfiguration;
            $this->redis_client = new Redis();
            $this->ping_count = 3;
        }

        /**
         * Attempts to determine if the Redis Server is online.
         *
         * @return bool Returns true if the Redis Server is online, false otherwise.
         */
        private function isConnected(): bool
        {
            if($this->redis_client === null)
            {
                return false;
            }

            try
            {
                // Ping every 3 calls
                if($this->ping_count >= 3)
                {
                    $this->redis_client->ping();
                    $this->ping_count = 0;
                }
                else
                {
                    ++$this->ping_count;
                }
            }
            catch(Exception $e)
            {
                unset($e);
                return false;
            }

            return true;
        }

        /**
         * Attempts to connect to the Redis Server
         *
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @return void
         */
        private function connect(): void
        {
            // Reconnect every 30 minutes
            if ($this->last_connect !== null && $this->last_connect < (time() - 1800))
            {
                Log::verbose(Utilities::getName(), sprintf('JobManager reconnecting to %s:%s (Last connect timeout)', $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                $this->disconnect();
            }

            if($this->isConnected())
            {
                return;
            }

            try
            {
                $this->redis_client->connect(
                    $this->server_configuration->getHost(),
                    $this->server_configuration->getPort()
                );

                if ($this->server_configuration->getPassword() !== null && $this->server_configuration->getPassword() !== '')
                {
                    $this->redis_client->auth($this->server_configuration->getPassword());
                }

                $this->redis_client->select($this->server_configuration->getDatabase());
                $this->last_connect = time();
                $this->redis_client->ping();
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Unable to connect to %s:%s, %s', $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('There was an unexpected error while trying to connect to %s:%s, %s', $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
        }

        /**
         * Disconnects from the Redis Server
         *
         * @return void
         */
        public function disconnect(): void
        {
            if($this->isConnected())
            {
                return;
            }

            Log::debug(Utilities::getName(), sprintf('JobManager disconnecting from %s:%s', $this->server_configuration->getHost(), $this->server_configuration->getPort()));

            try
            {
                $this->redis_client->close();
            }
            catch(Exception $e)
            {
                Log::warning(Utilities::getName(),sprintf('JobManager could not disconnect safely from %s:%s', $this->server_configuration->getHost(), $this->server_configuration->getPort()), $e);
            }
            finally
            {
                $this->redis_client = null;
            }
        }

        /**
         * Returns the Redis Client, or attempts to connect to the Redis Server if not connected.
         *
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @return Redis Returns the Redis Client
         */
        private function getClient(): Redis
        {
            $this->connect();
            return $this->redis_client;
        }

        /**
         * Pushes a JobPacket to the Redis server, this will take the JobPacket, push it as a hash to the server
         * then push the JobPacket ID to the channel list specified in the JobPacket, this will allow workers to
         * pull the JobPacket from the server and process it.
         *
         * @param JobPacket $jobPacket The JobPacket to push to the Redis server
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @return void
         */
        public function pushJob(JobPacket $jobPacket): void
        {
            try
            {
                Log::debug(Utilities::getName(), sprintf('JobManager pushing job %s to %s:%s', $jobPacket->getId(), $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                $this->getClient()->hMSet($jobPacket->getId(), $jobPacket->toArray());
                $this->getClient()->rPush(sprintf('ch%s', $jobPacket->getChannel()), $jobPacket->getId());
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('There was an error while trying to push job %s to %s:%s, %s', $jobPacket->getId(), $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('There was an unexpected error while trying to push job %s to %s:%s, %s', $jobPacket->getId(), $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
        }

        /**
         * Attempts to get a specific job from the Redis Server, if the requested JobPacket does not exist on the
         * server then a JobNotFoundException will be thrown.
         *
         * @param JobPacket|string $job_id The JobPacket or JobPacket ID to get from the Redis Server
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @throws JobNotFoundException Thrown if the JobPacket does not exist on the Redis Server
         * @return JobPacket Returns the JobPacket from the Redis Server
         */
        public function getJob(JobPacket|string $job_id): JobPacket
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            Log::debug(Utilities::getName(), sprintf('JobManager getting job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));

            try
            {
                if(!$this->getClient()->exists($job_id))
                {
                    throw new JobNotFoundException(sprintf('Job %s does not exist in %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                }

                return new JobPacket($this->getClient()->hGetAll($job_id));

            }
            catch(JobNotFoundException $e)
            {
                throw $e;
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Client threw an error while trying to get job %s from %s:%s, %s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('There was an unexpected error while trying to get job %s from %s:%s, %s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
        }

        /**
         * Returns the status of a job, optionally if $wait_for is set to an array of statuses then the method will
         * wait until the job status is one of the statuses in the array before returning. In this case the $timeout
         * parameter will be used to determine how long to wait for the job status to change.
         *
         * Returns the status of the job as an integer, the integer will be one of the following:
         *
         *  - JobStatus::REJECTED = 50
         *  - JobStatus::FAILED = 40
         *  - JobStatus::FINISHED = 30
         *  - JobStatus::PROCESSING = 20
         *  - JobStatus::WAITING = 10
         *
         * @param JobPacket|string $job_id The JobPacket or JobPacket ID to get the status of
         * @param array|null $wait_for Optional. An array of statuses to wait for before returning the status
         * @param int $timeout Optional. The number of seconds to wait for the status to change if $wait_for is set
         * @return int Returns the status of the job as an integer, see JobStatus for the integer values of the statuses
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @throws JobNotFoundException Thrown if the JobPacket does not exist on the Redis Server
         * @throws TimeoutException Thrown if the timeout is reached before the job status changes to one of the statuses in $wait_for
         * @throws JobNotFoundException
         * @see JobStatus for the integer values of the statuses
         */
        public function getJobStatus(JobPacket|string $job_id, ?array $wait_for=null, int $timeout=0): int
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            if($wait_for !== null)
            {
                $start_time = time();
                Log::debug(Utilities::getName(), sprintf('JobManager waiting for job %s to be one of the following statuses: %s', $job_id, implode(', ', $wait_for)));
                while(true)
                {
                    $job = $this->getJob($job_id);

                    if(in_array($job->getStatus(), $wait_for, true))
                    {
                        return $job->getStatus();
                    }

                    if($timeout > 0 && (time() - $start_time) > $timeout)
                    {
                        throw new TimeoutException(sprintf('Timed out waiting for job %s to be one of the following statuses: %s', $job_id, implode(', ', $wait_for)));
                    }

                    Utilities::ausleep(100000, static function(){
                        tm::monitor(-1);
                    });
                }
            }

            Log::debug(Utilities::getName(), sprintf('JobManager getting status of job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));

            try
            {
                if(!$this->getClient()->exists($job_id))
                {
                    throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                }

                return (int)$this->getClient()->hGet($job_id, 'status');
            }
            catch(JobNotFoundException $e)
            {
                throw $e;
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Client threw an error while trying to get job %s from %s:%s, %s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('There was an unexpected error while trying to get job %s from %s:%s, %s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
        }

        /**
         * Listens on a return channel for a returned job, optionally with a timeout. If the timeout is reached before
         * a job is returned then a TimeoutException will be thrown.
         *
         * @param string $return_channel The return channel to listen on
         * @param int $timeout Optional. The number of seconds to wait for a job to be returned
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @throws TimeoutException Thrown if the timeout is reached before a job is returned
         * @return string Returns the returned job ID
         */
        public function listenReturnChannel(string $return_channel, int $timeout=0): string
        {
            try
            {
                if($timeout < 0)
                {
                    Log::debug(Utilities::getName(), sprintf('Listening on return channel (LPOP) %s on %s:%s', $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                    $job_packet = $this->getClient()->lPop($return_channel);
                }
                else
                {
                    Log::debug(Utilities::getName(), sprintf('Listening on return channel %s on %s:%s with a timeout of %s seconds', $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $timeout));

                    $start_time = time();
                    while(true)
                    {
                        $job_packet = $this->getClient()->lPop($return_channel);

                        if(!is_bool($job_packet) && $job_packet !== null)
                        {
                            break;
                        }

                        if($timeout > 0 && (time() - $start_time) > $timeout)
                        {
                            throw new TimeoutException(sprintf('Timed out listening on return channel %s on %s:%s', $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                        }

                        Utilities::ausleep(10000, static function(){
                            tm::monitor(-1);
                        });
                    }
                }
            }
            catch(TimeoutException $e)
            {
                throw $e;
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Client threw an error while trying to get job from return channel %s on %s:%s, %s', $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('Could not get job from return channel %s on %s:%s', $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort()), $e);
            }

            if(is_bool($job_packet) && $job_packet === false)
            {
                throw new TimeoutException(sprintf('Could not get job from return channel %s on %s:%s', $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
            }

            Log::debug(Utilities::getName(), sprintf('Got job %s from return channel %s on %s:%s', $job_packet, $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
            return $job_packet;
        }

        /**
         * Pushes a job back to the return channel. This is useful if a job is returned to the return channel, but the
         * client does not want to process it yet.
         *
         * @param JobPacket|string $job_id
         * @param string $return_channel
         * @return void
         * @throws ConnectionException
         * @throws JobManagerException
         */
        public function pushbackJob(JobPacket|string $job_id, string $return_channel): void
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            Log::debug(Utilities::getName(), sprintf('Pushing job %s back to return channel %s on %s:%s', $job_id, $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort()));

            try
            {
                if(!$this->getClient()->exists($job_id))
                {
                    throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                }

                $this->getClient()->rPush($return_channel, $job_id);
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Client threw an error while trying to get job from return channel %s on %s:%s, %s', $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('Could not get job from return channel %s on %s:%s', $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort()), $e);
            }

        }

        /**
         * Returns the unserialized return value of a job.
         *
         * @param JobPacket|string $job_id The JobPacket or job ID to get the return value of
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @throws JobNotFoundException Thrown if the JobPacket does not exist on the server
         * @return mixed Returns the unserialized return value of the job
         */
        public function getJobResult(JobPacket|string $job_id): mixed
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            Log::debug(Utilities::getName(), sprintf('Getting return value of job %s on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));

            try
            {
                if(!$this->getClient()->exists($job_id))
                {
                    throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                }

                return unserialize($this->getClient()->hGet($job_id, 'return'), ['allowed_classes' => true]);
            }
            catch(JobNotFoundException $e)
            {
                throw $e;
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Client threw an error while trying to get job %s from %s:%s, %s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('Could not get job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()),  $e);
            }
        }

        /**
         * Attempts to unserialize and return the exception thrown by a job, if the exception cannot be unserialized,
         * a generic Exception is returned explaining that the exception could not be unserialized and what it got
         * instead.
         *
         * @param JobPacket|string $job_id The JobPacket or job ID to get the exception to
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @throws JobNotFoundException Thrown if the JobPacket does not exist on the server
         * @return Throwable Returns the unserialized exception thrown by the job
         */
        public function getJobException(JobPacket|string $job_id): Throwable
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            Log::debug(Utilities::getName(), sprintf('Getting exception of job %s on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));

            try
            {
                if(!$this->getClient()->exists($job_id))
                {
                    throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                }

                $exception = unserialize($this->getClient()->hGet($job_id, 'exception'), ['allowed_classes'=>true]);

                if($exception instanceof Throwable)
                {
                    return $exception;
                }

                if($exception === false)
                {
                    return new Exception(sprintf('Job %s threw an exception, but the exception could not be unserialized.', $job_id));
                }

                return new Exception(sprintf('Job %s threw an exception, but the exception is not an instance of Throwable, got \'%s\'', $job_id, gettype($exception)));
            }
            catch(JobNotFoundException $e)
            {
                throw $e;
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Client threw an error while trying to get job %s from %s:%s, %s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('Could not get job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), $e);
            }
        }

        /**
         * Attempts to claim an existing JobPacket, returns True if the worker's claim was successful.
         * Returns False if the JobPacket does not exist, or if the JobPacket is already claimed.
         *
         * This also automatically sets the JobPacket's status to 'processing' and sets the worker ID
         * to the worker claiming the job.
         *
         * @param JobPacket|string $job_id The JobPacket or job ID to claim
         * @param WorkerConfiguration|string $worker_id The WorkerConfiguration or worker ID to claim the job with
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @return bool Returns True if the worker successfully claimed the job, False otherwise
         */
        public function claimJob(JobPacket|string $job_id, WorkerConfiguration|string $worker_id): bool
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            if($worker_id instanceof WorkerConfiguration)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $worker_id = $worker_id->getWorkerId();
            }

            Log::debug(Utilities::getName(), sprintf('Attempting to claim job %s on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));

            try
            {
                // Check if the job exists
                if(!$this->getClient()->exists($job_id))
                {
                    return false;
                }

                // Attempt to claim the job
                $this->getClient()->hSet($job_id, 'worker_id', $worker_id);

                // Verify that the job was claimed
                if($this->getClient()->hGet($job_id, 'worker_id') !== $worker_id)
                {
                    Log::warning(Utilities::getName(), sprintf('Job %s on %s:%s was already claimed by %s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $this->getClient()->hGet($job_id, 'worker_id')));
                    return false;
                }

                // Set the job status to processing
                $this->getClient()->hSet($job_id, 'status', JobStatus::PROCESSING);
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Client threw an error while trying to claim job %s from %s:%s, %s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('Could not claim job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), $e);
            }

            return true;
        }

        /**
         * Rejects a job and pushes it back to the channel, allowing another worker to claim it.
         *
         * @param JobPacket|string $job_id The JobPacket or job ID to reject
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @throws JobNotFoundException Thrown if the JobPacket does not exist on the server
         * @return void
         */
        public function rejectJob(JobPacket|string $job_id): void
        {
            $channel_id = null;
            $was_job_packet = false;

            if($job_id instanceof JobPacket)
            {
                // Providing a JobPacket allows us to get the channel_id and avoid an extra call to the server
                $channel_id = $job_id->getChannel();
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
                $was_job_packet = true;
            }

            try
            {
                if(!$this->getClient()->exists($job_id))
                {
                    throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                }

                Log::debug(Utilities::getName(), sprintf('Rejecting job %s', $job_id));

                // Mark as rejected, clear worker_id
                $this->getClient()->hSet($job_id, 'worker_id', null);

                if($channel_id === null && !$was_job_packet)
                {
                    // Get the channel_id if we don't have it already
                    $channel_id = $this->getClient()->hGet($job_id, 'channel_id');
                }

                if($channel_id !== null)
                {
                    $this->getClient()->rPush(sprintf('ch%s', $channel_id), $job_id);
                }
            }
            catch(JobNotFoundException $e)
            {
                throw $e;
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Client threw an error while trying to reject job %s from %s:%s, %s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('Could not reject job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), $e);
            }
        }

        /**
         * Marks a job as finished, and sets the return value. This basically returns the job to the client.
         * If there is no return_channel set, the job is deleted from the server.
         *
         * @param JobPacket|string $job_id The JobPacket or job ID to return
         * @param mixed|null $return_value The return value to set on the job
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @throws JobNotFoundException Thrown if the JobPacket does not exist on the server
         * @return void
         */
        public function returnJob(JobPacket|string $job_id, mixed $return_value=null): void
        {
            $was_job_packet = false;

            if($job_id instanceof JobPacket)
            {
                $return_channel = $job_id->getReturnChannel();
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
                $was_job_packet = true;
            }

            Log::debug(Utilities::getName(), sprintf('Returning job %s', $job_id));

            try
            {
                if(!$this->getClient()->exists($job_id))
                {

                    throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                }

                if(!$was_job_packet)
                {
                    $return_channel = $this->getClient()->hGet($job_id, 'return_channel');
                }

                /** @noinspection PhpUndefinedVariableInspection */
                if($return_channel === null || $return_channel === '')
                {
                    Log::debug(Utilities::getName(), sprintf('No return channel set, deleting job %s', $job_id));
                    $this->getClient()->del($job_id);
                    return;
                }

                Log::debug(Utilities::getName(), sprintf('Pushing job %s to return channel %s', $job_id, $return_channel));
                $this->getClient()->hSet($job_id, 'return_value', serialize($return_value));
                $this->getClient()->hSet($job_id, 'status', JobStatus::FINISHED);
                $this->getClient()->rPush($this->getClient()->hGet($job_id, 'return_channel'), $job_id);
            }
            catch(JobNotFoundException $e)
            {
                throw $e;
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Client threw an error while trying to return job %s from %s:%s, %s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('Could not return job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), $e);
            }
        }

        /**
         * Sets the job as failed, and sets the exception that was thrown.
         *
         * @param JobPacket|string $job_id The JobPacket or job ID to return
         * @param Throwable $throwable The exception that was thrown
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @return void
         */
        public function returnException(JobPacket|string $job_id, Throwable $throwable): void
        {
            $was_job_packet = false;
            if($job_id instanceof JobPacket)
            {
                $return_channel = $job_id->getReturnChannel();
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
                $was_job_packet = true;
            }

            Log::debug(Utilities::getName(), sprintf('Returning exception for job %s', $job_id));

            try
            {
                if(!$this->getClient()->exists($job_id))
                {
                    throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                }

                if(!$was_job_packet)
                {
                    $return_channel = $this->getClient()->hGet($job_id, 'return_channel');
                }

                /** @noinspection PhpUndefinedVariableInspection */
                if($return_channel === null || $return_channel === '')
                {
                    Log::debug(Utilities::getName(), sprintf('No return channel set, deleting job %s', $job_id));
                    $this->getClient()->del($job_id);
                    return;
                }

                Log::debug(Utilities::getName(), sprintf('Pushing job %s to return channel %s', $job_id, $return_channel));
                $this->getClient()->hSet($job_id, 'exception', serialize($throwable));
                $this->getClient()->hSet($job_id, 'status', JobStatus::FAILED);
                $this->getClient()->rPush($return_channel, $job_id);
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Client threw an error while trying to return exception for job %s from %s:%s, %s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('Could not return exception for job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), $e);
            }
        }

        /**
         * Waits for a job to be available on the channel or multiple channels, attempts to claim it, and returns it.
         * If multiple channels are selected, the function will iterate through each channel until a job is found.
         *
         * @param WorkerConfiguration|string $worker_id The worker that is waiting for the job.
         * @param int|array $channel The channel (or channels) to wait for the job on.
         * @param int $timeout The timeout in seconds to wait for the job.
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @throws TimeoutException Thrown if the timeout is reached before a job is available
         * @return JobPacket The job that was claimed, or null if no job was available.
         */
        public function listenForJob(WorkerConfiguration|string $worker_id, int|array $channel=0, int $timeout=0): JobPacket
        {
            if($worker_id instanceof WorkerConfiguration)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $worker_id = $worker_id->getWorkerId();
            }

            $channels = [];
            if(is_array($channel))
            {
                foreach($channel as $id)
                {
                    $channels[] = sprintf('ch%s', $id);
                }
            }
            else
            {
                $channels[] = sprintf('ch%s', $channel);
            }

            Log::debug(Utilities::getName(), sprintf('Waiting for job on channels %s', implode(', ', $channels)));

            try
            {
                while(true)
                {
                    if($timeout < 0)
                    {
                        $job_id = $this->getClient()->blPop($channels, 1);
                    }
                    else
                    {
                        $job_id = $this->getClient()->blPop($channels, $timeout);
                    }

                    if($job_id !== false && $this->claimJob($job_id[1], $worker_id))
                    {
                        Log::debug(Utilities::getName(), sprintf('Claimed job %s', $job_id[1]));
                        return new JobPacket($this->getClient()->hGetAll($job_id[1]));
                    }

                    if($job_id !== false)
                    {
                        Log::debug(Utilities::getName(), sprintf('Could not claim job %s', $job_id[1]));
                    }

                    if($timeout < 0)
                    {
                        throw new TimeoutException(sprintf('Timeout exceeded while waiting for job on %s:%s', $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                    }

                    if($timeout === 0 && $job_id === null)
                    {
                        throw new TimeoutException(sprintf('Timeout exceeded while waiting for job on %s:%s', $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                    }
                }

            }
            catch(TimeoutException $e)
            {
                throw $e;
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Client threw an error while trying to listen for job on %s:%s, %s', $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('Could not listen for job on %s:%s', $this->server_configuration->getHost(), $this->server_configuration->getPort()), $e);
            }
        }

        /**
         * Drops a job from the server entirely, if the job doesn't exist then nothing happens.
         *
         * @param JobPacket|string $job_id The job to drop.
         * @throws ConnectionException Thrown if there is a connection issue with the server
         * @throws JobManagerException Thrown if there is an issue with the JobManager
         * @return void
         */
        public function dropJob(JobPacket|string $job_id): void
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            Log::debug(Utilities::getName(), sprintf('Dropping job %s', $job_id));

            try
            {
                if(!$this->getClient()->exists($job_id))
                {
                    return;
                }

                $this->getClient()->del($job_id);
            }
            catch(RedisException $e)
            {
                throw new ConnectionException(sprintf('Client threw an error while trying to drop job %s from %s:%s, %s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $e->getMessage()), $e);
            }
            catch(Exception $e)
            {
                throw new JobManagerException(sprintf('Could not drop job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), $e);
            }
        }
    }