<?php

    /** @noinspection PhpMissingFieldTypeInspection */

    namespace TamerLib\Classes;

    use LogLib\Abstracts\LevelType;
    use LogLib\Log;
    use Redis;
    use RedisException;
    use Symfony\Component\Process\Process;
    use TamerLib\Exceptions\ServerException;
    use TamerLib\Objects\ServerConfiguration;
    use TamerLib\tm;

    class RedisServer
    {
        /**
         * @var string
         */
        private $cmd;

        /**
         * The server configuration
         *
         * @var ServerConfiguration
         */
        private $configuration;

        /**
         * @var Process|null
         */
        private $server_process;

        /**
         * RedisServer constructor.
         *
         * @param ServerConfiguration $configuration
         * @param string $cmd
         */
        public function __construct(ServerConfiguration $configuration, string $cmd='redis-server')
        {
            $this->configuration = $configuration;
            $this->cmd = $cmd;
        }

        /**
         * Determines if the Redis server is running.
         *
         * @return bool
         */
        public function isRunning(): bool
        {
            if(is_null($this->server_process))
            {
                return false;
            }

            return $this->server_process->isRunning();
        }

        /**
         * Starts the Redis server.
         *
         * @param int $timeout
         * @return bool
         * @throws ServerException
         */
        public function start(int $timeout=60): bool
        {
            if($this->isRunning())
            {
                return true;
            }

            Log::verbose(Utilities::getName(), 'Starting server on port ' . $this->configuration->getPort() . '.');

            $this->server_process = new Process([
                $this->cmd, __DIR__ . DIRECTORY_SEPARATOR . 'redis.conf',
                '--port', $this->configuration->getPort(),
                '--loglevel', match (\LogLib\Classes\Utilities::getLogLevel())
                {
                    LevelType::Warning, LevelType::Error => 'warning',
                    LevelType::Verbose => 'verbose',
                    LevelType::Debug => 'debug',
                    default => 'notice',
                }
            ]);

            $this->server_process->start();

            // Use a redis client and ping the server until it responds.
            $redis_client = new Redis();
            $timeout_counter = 0;

            while(true)
            {
                tm::monitor(-1);
                if($timeout_counter >= $timeout)
                {
                    throw new ServerException('Redis server failed to start within ' . $timeout . ' seconds.');
                }

                try
                {
                    if($redis_client->isConnected())
                    {
                        break;
                    }

                    $redis_client->connect($this->configuration->getHost(), $this->configuration->getPort());
                }
                catch (RedisException $e)
                {
                    // Do nothing.
                    unset($e);
                }
                finally
                {
                    sleep(1);
                    $timeout_counter++;
                }
            }
            
            Log::verbose(Utilities::getName(), sprintf('Server listening on %s:%s.', $this->configuration->getHost(), $this->configuration->getPort()));
            return true;
        }

        /**
         * Stops the Redis server.
         *
         * @return bool
         */
        public function stop(): bool
        {
            if(!$this->isRunning())
            {
                return true;
            }

            $this->server_process->stop();
            Log::verbose(Utilities::getName(), sprintf('Server stopped on %s:%s.', $this->configuration->getHost(), $this->configuration->getPort()));
            return true;
        }

        /**
         * Monitors the server process and outputs the latest output.
         *
         * @param int $timeout
         * @return void
         * @throws ServerException
         */
        public function monitor(int $timeout=0): void
        {
            $start_time = time();
            while(true)
            {
                print(Utilities::getLatestOutput($this->server_process));

                if(!$this->isRunning())
                {
                    Log::warning(Utilities::getName(), sprintf('Server on %s:%s is not running, restarting.', $this->configuration->getHost(), $this->configuration->getPort()));
                    $this->start();
                }

                if($timeout < 0)
                {
                    return;
                }

                if($timeout > 0 && (time() - $start_time) >= $timeout)
                {
                    return;
                }
            }

        }

        /**
         * Terminates the Redis server.
         */
        public function __destruct()
        {
            $this->stop();
        }
    }