<?php

    /** @noinspection PhpMissingFieldTypeInspection */

    namespace TamerLib\Classes;

    use Exception;
    use Symfony\Component\Process\Process;
    use TamerLib\Enums\TamerMode;
    use TamerLib\Exceptions\NoAvailablePortException;
    use TamerLib\tm;
    use Throwable;

    class Utilities
    {
        /**
         * @var int|null
         */
        private static $pid;

        /**
         * Attempts to find an available port in the given range. If no port is available,
         * a NoAvailablePortException is thrown.
         *
         * @param string $host
         * @param int $start
         * @param int $end
         * @return int
         * @throws NoAvailablePortException
         */
        public static function getAvailablePort(string $host='127.0.0.1', int $start=1024, int $end=65535): int
        {
            $range = range($start, $end);
            shuffle($range);
            foreach ($range as $port)
            {
                $connection = @stream_socket_client('tcp://' . $host . ':' . $port);
                if (is_resource($connection))
                {
                    fclose($connection);
                }
                else
                {
                    return $port;
                }
            }

            throw new NoAvailablePortException('No available port found in range ' . $start . ' to ' . $end . '.');
        }

        /**
         * Returns a randomly generated string of the given length.
         *
         * @param int $length
         * @return string
         * @throws Exception
         */
        public static function generateRandomString(int $length=8): string
        {
            $characters = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ';
            $characters_length = strlen($characters);
            $random_string = '';
            for ($i = 0; $i < $length; $i++)
            {
                $random_string .= $characters[random_int(0, $characters_length - 1)];
            }
            return $random_string;
        }

        /**
         * Returns an array representation of a throwable exception.
         *
         * @param Throwable $throwable
         * @param bool $recursive
         * @return array
         */
        public static function throwableToArray(Throwable $throwable, bool $recursive=true): array
        {
            $array = [
                'message' => $throwable->getMessage(),
                'code' => $throwable->getCode(),
                'file' => $throwable->getFile(),
                'line' => $throwable->getLine(),
                'trace' => $throwable->getTrace(),
            ];

            if($recursive && $throwable->getPrevious() instanceof Throwable)
            {
                $array['previous'] = self::throwableToArray($throwable->getPrevious());
            }

            return $array;
        }

        /**
         * Returns the current name of the process for logging purposes.
         * If the current mode is worker, the process name will be
         * net.nosial.tamerlib:<pid>. Otherwise, the process name will be
         * net.nosial.tamerlib.
         *
         * @return string
         */
        public static function getName(): string
        {
            if(tm::getMode() === TamerMode::WORKER)
            {
                if(self::$pid === null)
                {
                    // Very unlikely that getmypid() will return 0.
                    /** @noinspection ProperNullCoalescingOperatorUsageInspection */
                    self::$pid = getmypid() ?? (string)null;
                }

                return sprintf('net.nosial.tamerlib:%s', self::$pid);
            }

            return 'net.nosial.tamerlib';
        }

        /**
         * Returns the latest output from the given process either from the output or error output.
         * If no output is available, an empty string is returned.
         *
         * @param Process $process
         * @return string
         */
        public static function getLatestOutput(Process $process): string
        {
            $output = $process->getIncrementalOutput();
            $error_output = $process->getIncrementalErrorOutput();

            if (!empty($error_output))
            {
                $output .= PHP_EOL . $error_output;
            }

            return empty($output) ? (string)null : $output;
        }

        /**
         * Sleeps for a given number of seconds. If a callback is provided, it will be called
         * every second until the sleep is complete.
         *
         * @param int $seconds
         * @param callable|null $callback
         * @return void
         */
        public static function asleep(int $seconds, ?callable $callback=null): void
        {
            $start = time();
            while(time() - $start < $seconds)
            {
                if($callback !== null)
                {
                    $callback();
                }
                sleep(1);
            }
        }

        /**
         * Sleeps for a given number of microseconds. If a callback is provided, it will be called
         * every millisecond until the sleep is complete.
         *
         * @param int $microseconds
         * @param callable|null $callback
         * @return void
         */
        public static function ausleep(int $microseconds, ?callable $callback = null): void
        {
            $start = microtime(true);
            $elapsed = 0;

            while ($elapsed < $microseconds)
            {
                if ($callback !== null) {
                    $callback();
                }

                usleep(1000);
                $elapsed = (microtime(true) - $start) * 1000000;
            }
        }
    }